<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace DistributeModule\Lib;


use App\Lib\PluginExportTasks;
use App\Model\Entity\Hostgroup;
use App\Model\Entity\Servicegroup;
use App\Model\Entity\Timeperiod;
use App\Model\Table\HostgroupsTable;
use App\Model\Table\HostsTable;
use App\Model\Table\ServicegroupsTable;
use App\Model\Table\ServicesTable;
use App\Model\Table\TimeperiodsTable;
use Cake\Core\Configure;
use Cake\Filesystem\Folder;
use Cake\ORM\Query;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use DistributeModule\Model\Entity\Satellite;
use itnovum\openITCOCKPIT\Satellite\MappingRepository;

class ExportTasks implements PluginExportTasks {

    public function beforeExport(): bool {
        $this->createMapping();
        return true;
    }

    public function afterExport(): bool {
        $this->copyDefaultConfig();
        return true;
    }

    private function createMapping() {
        Configure::load('nagios');
        $conf = Configure::read('nagios.export');

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
        /** @var ServicesTable $ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');
        /** @var TimeperiodsTable $TimeperiodsTable */
        $TimeperiodsTable = TableRegistry::getTableLocator()->get('Timeperiods');
        /** @var \DistributeModule\Model\Table\SatellitesTable $SatellitesTable */
        $SatellitesTable = TableRegistry::getTableLocator()->get('DistributeModule.Satellites');
        /** @var HostgroupsTable $HostgroupsTable */
        $HostgroupsTable = TableRegistry::getTableLocator()->get('Hostgroups');
        /** @var ServicegroupsTable $ServicegroupsTable */
        $ServicegroupsTable = TableRegistry::getTableLocator()->get('Servicegroups');

        $satellites = $SatellitesTable->find()->all();
        if (empty($satellites)) {
            return true;
        }
        $hostgroups = $HostgroupsTable->find()->contain('Containers')->all();
        $servicegroups = $ServicegroupsTable->find()->contain('Containers')->all();

        foreach ($satellites as $satellite) {
            /** @var Satellite $satellite */
            if (!is_dir($conf['satellite_path'] . $satellite->get('id'))) {
                mkdir($conf['satellite_path'] . $satellite->get('id'));
            }

            $hostIds = [];
            $serviceIds = [];

            $MappingRepository = new MappingRepository();
            $hosts = $HostsTable->find()
                ->select([
                    'Hosts.id',
                    'Hosts.uuid',
                    'Hosts.name'
                ])
                ->where([
                    'Hosts.satellite_id' => $satellite->get('id'),
                    'Hosts.disabled'     => 0
                ])
                ->disableHydration()
                ->all();

            if (empty($hosts)) {
                $hosts = [];
            }
            foreach ($hosts as $host) {
                $MappingRepository->addItem($host['uuid'], $host['name'], OBJECT_HOST);
                $hostIds[$host['id']] = $host['uuid'];

                $services = $ServicesTable->find()
                    ->select([
                        'Services.id',
                        'Services.name',
                        'Services.uuid',
                        'Services.servicetemplate_id'
                    ])
                    ->contain([
                        'Servicetemplates' => function (Query $query) {
                            $query
                                ->disableAutoFields()
                                ->select([
                                    'id',
                                    'name'
                                ]);
                            return $query;
                        }
                    ])
                    ->where([
                        'Services.host_id'  => $host['id'],
                        'Services.disabled' => 0
                    ])
                    ->disableHydration()
                    ->all();


                if (empty($services)) {
                    $services = [];
                }
                foreach ($services as $service) {
                    $serviceIds[$service['id']] = $service['uuid'];

                    $serviceName = $service['name'];
                    if ($serviceName === null || $serviceName === '') {
                        $serviceName = $service['servicetemplate']['name'];
                    }
                    $MappingRepository->addItem($service['uuid'], $serviceName, OBJECT_SERVICE);
                }
            }

            $timeperiods = $TimeperiodsTable->find()->all();

            foreach ($timeperiods as $timeperiod) {
                /** @var Timeperiod $timeperiod */
                $MappingRepository->addItem(
                    $timeperiod->get('uuid'),
                    $timeperiod->get('name'),
                    OBJECT_TIMEPERIOD
                );
            }

            if (!empty($hostIds) && !empty($serviceIds)) {
                foreach ($hostgroups as $hostgroup) {
                    /** @var Hostgroup $hostgroup */
                    $hostgroupHosts = $HostgroupsTable->getHostsByHostgroupIdAndSatelliteId(
                        $hostgroup->get('id'),
                        $satellite['id']
                    );
                    if (!empty($hostgroupHosts)) {
                        $MappingRepository->addItem(
                            $hostgroup->get('uuid'),
                            $hostgroup->get('container')->get('name'),
                            OBJECT_HOSTGROUP
                        );

                        $MappingRepository->addHostgroup(
                            $hostgroup->get('id'),
                            $hostgroup->get('uuid'),
                            $hostgroup->get('container')->get('name'),
                            $hostgroup->get('description'),
                            Hash::extract($hostgroupHosts, '{n}.uuid')
                        );

                    }
                }

                foreach ($servicegroups as $servicegroup) {
                    /** @var Servicegroup $servicegroup */
                    $servicegroupServices = $ServicegroupsTable->getServicesByServicegroupIdAndSatelliteId(
                        $servicegroup->get('id'),
                        $satellite['id']
                    );
                    if (!empty($servicegroupServices)) {
                        $MappingRepository->addItem(
                            $servicegroup->get('uuid'),
                            $servicegroup->get('container')->get('name'),
                            OBJECT_SERVICEGROUP
                        );

                        $MappingRepository->addServicegroup(
                            $servicegroup->get('id'),
                            $servicegroup->get('uuid'),
                            $servicegroup->get('container')->get('name'),
                            $servicegroup->get('description'),
                            Hash::extract($servicegroupServices, '{n}.uuid')
                        );
                    }
                }
            }

            $MappingRepository->toFile($conf['satellite_path'] . $satellite->get('id') . DS . 'mapping.json');
        }
    }

    private function copyDefaultConfig() {
        Configure::load('nagios');
        $conf = Configure::read('nagios.export');

        /** @var \DistributeModule\Model\Table\SatellitesTable $SatellitesTable */
        $SatellitesTable = TableRegistry::getTableLocator()->get('DistributeModule.Satellites');

        $satellites = $SatellitesTable->find()->all();
        if (empty($satellites)) {
            return true;
        }

        $dir2copy = [
            'commands',
            'contactgroups',
            'contacts',
            'hosttemplates',
            'servicetemplates',
        ];

        $defaultFolder = new Folder(Configure::read('nagios.basepath') . Configure::read('nagios.etc') . Configure::read('nagios.export.config') . 'defaults/');
        $rolloutFolder = new Folder($conf['rollout']);

        foreach ($satellites as $satellite) {
            /** @var Satellite $satellite */
            if (is_dir($conf['satellite_path'] . $satellite->get('id'))) {

                //Copy default settings like generic host, timeperiod and stuff
                $defaultFolder->copy($conf['satellite_path'] . $satellite->get('id') . DS . $conf['config'] . 'defaults/');

                foreach ($dir2copy as $dir) {
                    if (is_dir($conf['path'] . $conf[$dir])) {
                        $sourceFolder = new Folder($conf['path'] . $conf[$dir]);
                        if (!is_dir($conf['satellite_path'] . $satellite->get('id') . DS . $conf[$dir])) {
                            mkdir($conf['satellite_path'] . $satellite->get('id') . DS . $conf[$dir]);
                        }
                        $sourceFolder->copy($conf['satellite_path'] . $satellite->get('id') . DS . $conf[$dir]);
                    }
                }
            }

            //Copy files from rolloit folder to sat config folder
            $rolloutFolder->copy($conf['satellite_path'] . $satellite->get('id') . DS);

        }
    }

}
