<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace DistributeModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Lib\Traits\PluginManagerTableTrait;
use App\Model\Table\ContainersTable;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use DistributeModule\Lib\SatSyncMethods;
use itnovum\openITCOCKPIT\Core\FileDebugger;
use itnovum\openITCOCKPIT\Core\SystemdowntimesConditions;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\SatelliteFilter;


/**
 * Satellites Model
 *
 * @property \App\Model\Table\ContainersTable|\Cake\ORM\Association\BelongsTo $Containers
 * @property \App\Model\Table\HostsTable|\Cake\ORM\Association\HasMany $Hosts
 *
 * @method \DistributeModule\Model\Entity\Satellite get($primaryKey, $options = [])
 * @method \DistributeModule\Model\Entity\Satellite newEntity($data = null, array $options = [])
 * @method \DistributeModule\Model\Entity\Satellite[] newEntities(array $data, array $options = [])
 * @method \DistributeModule\Model\Entity\Satellite|bool save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \DistributeModule\Model\Entity\Satellite saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \DistributeModule\Model\Entity\Satellite patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \DistributeModule\Model\Entity\Satellite[] patchEntities($entities, array $data, array $options = [])
 * @method \DistributeModule\Model\Entity\Satellite findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TimestampBehavior
 */
class SatellitesTable extends Table {

    use PaginationAndScrollIndexTrait;
    use PluginManagerTableTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('satellites');
        $this->setDisplayField('name');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Containers', [
            'foreignKey' => 'container_id',
            'joinType'   => 'INNER',
            'className'  => 'DistributeModule.Containers',
        ]);

        $this->hasMany('Hosts', [
            'foreignKey' => 'satellite_id',
            'className'  => 'DistributeModule.Hosts',
        ]);

        $this->hasOne('SatelliteStatus', [
            'joinType'  => 'LEFT',
            'className' => 'DistributeModule.SatelliteStatus',
            'dependent' => false //ITC-2350
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $SatSyncMethods = new SatSyncMethods();

        $validator
            ->integer('id')
            ->allowEmptyString('id', 'create');

        $validator
            ->scalar('name')
            ->maxLength('name', 255)
            ->requirePresence('name', 'create')
            ->notEmptyString('name');

        $validator
            ->scalar('address')
            ->maxLength('address', 128)
            ->requirePresence('address', 'create')
            ->notEmptyString('address');

        $validator
            ->integer('container_id')
            ->requirePresence('container_id', 'create')
            ->allowEmptyString('container_id', null, false)
            ->greaterThanOrEqual('container_id', 1);

        $validator
            ->scalar('login')
            ->maxLength('login', 255)
            ->allowEmptyString('login', null, function ($context) {
                if ($context['data']['sync_method'] === 'https_pull' || $context['data']['sync_method'] === 'https_push') {
                    return true;
                }
                return !empty($context['data']['login']);
            });

        $validator
            ->scalar('private_key_path')
            ->maxLength('private_key_path', 255)
            ->requirePresence('private_key_path', 'create')
            ->allowEmptyString('private_key_path', null, function ($context) {
                if ($context['data']['sync_method'] === 'https_pull' || $context['data']['sync_method'] === 'https_push') {
                    return true;
                }
                return !empty($context['data']['private_key_path']);
            });

        $validator
            ->integer('sync_instance')
            ->allowEmptyString('sync_instance');

        $validator
            ->scalar('timezone')
            ->maxLength('timezone', 128)
            ->notEmptyString('timezone');

        $validator
            ->scalar('sync_method')
            ->maxLength('sync_method', 255)
            ->notEmptyString('sync_method')
            ->inList('sync_method', array_keys($SatSyncMethods->getSyncMethods()), __('Synchronisation method is out of bounds: [{0}]', implode(', ', array_keys($SatSyncMethods->getSyncMethods()))));

        $validator
            ->scalar('port')
            ->integer('port')
            ->range('port', [0, 65535], __('The port numbers in the range from 0 to 65535'))
            ->allowEmptyString('port', null, function ($context) {
                if ($context['data']['sync_method'] === 'https_pull' || $context['data']['sync_method'] === 'https_push') {
                    return true;
                }
                return !empty($context['data']['port']);
            });

        $validator
            ->scalar('remote_port')
            ->integer('remote_port')
            ->range('remote_port', [0, 65535], __('The port numbers in the range from 0 to 65535'))
            ->allowEmptyString('remote_port', null, function ($context) {
                if ($context['data']['sync_method'] === 'https_pull' || $context['data']['sync_method'] === 'https_push') {
                    return true;
                }
                return !empty($context['data']['remote_port']);
            });

        $validator
            ->scalar('interval')
            ->allowEmptyString('interval', null, function ($context) {
                if ($context['data']['sync_method'] === 'ssh' || $context['data']['sync_method'] === 'https_push') {
                    return true;
                }
                return !empty($context['data']['interval']);
            })
            ->integer('interval')
            ->greaterThan('interval', 0);

        $validator
            ->scalar('timeout')
            ->allowEmptyString('timeout', null, function ($context) {
                if ($context['data']['sync_method'] === 'ssh') {
                    return true;
                }
                return !empty($context['data']['timeout']);
            })
            ->integer('timeout')
            ->greaterThan('timeout', 0);

        $validator
            ->allowEmptyString('url', null, function ($context) {
                if ($context['data']['sync_method'] === 'ssh' || $context['data']['sync_method'] === 'https_push') {
                    return true;
                }
                return !empty($context['data']['address'] . $context['data']['url']);
            })
            ->urlWithProtocol('url', __('The entered URL is invalid'), function ($context) {
                if ($context['data']['sync_method'] === 'ssh' || $context['data']['sync_method'] === 'https_push') {
                    return false;
                }
                return true;
            });

        $validator
            ->allowEmptyString('proxy_url', null, function ($context) {
                if ($context['data']['sync_method'] === 'ssh' || $context['data']['sync_method'] === 'https_push') {
                    return true;
                }
                if ($context['data']['use_proxy'] === 0) {
                    return true;
                }
                return !empty($context['data']['proxy_url']);
            })
            ->urlWithProtocol('proxy_url', __('The entered Proxy URL is invalid'));

        $validator
            ->allowEmptyString('api_key', null, function ($context) {
                if ($context['data']['sync_method'] === 'ssh') {
                    return true;
                }
                return !empty($context['data']['api_key']);
            });

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->isUnique(['name']));
        $rules->add($rules->existsIn(['container_id'], 'Containers'));

        return $rules;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Satellites.id' => $id]);
    }

    public function getSatellitesAsList($MY_RIGHTS = []) {
        $query = $this->find('list');
        if (!empty($MY_RIGHTS)) {
            $query->where([
                'Satellites.container_id IN' => $MY_RIGHTS
            ]);
        }
        $query->order([
            'Satellites.name' => 'ASC'
        ]);
        return $query->toArray();
    }

    /**
     * @param SatelliteFilter $SatelliteFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getSatellitesIndex(SatelliteFilter $SatelliteFilter, $PaginateOMat = null, $MY_RIGHTS = []) {
        $query = $this->find('all')
            ->where($SatelliteFilter->indexFilter())
            ->order($SatelliteFilter->getOrderForPaginator('Satellites.name', 'asc'));

        if (!empty($MY_RIGHTS)) {
            $query->andWhere([
                'Satellites.container_id IN' => $MY_RIGHTS
            ]);
        }

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param SatelliteFilter $SatelliteFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getSatellitesStatus(SatelliteFilter $SatelliteFilter, $PaginateOMat = null, $MY_RIGHTS = []) {
        $where = $SatelliteFilter->statusFilter();
        $having = [];
        if (isset($where['status IN'])) {
            $having['status IN'] = $where['status IN'];
            unset($where['status IN']);
        }

        $query = $this->find('all');
        $query->select([
            'Satellites.id',
            'Satellites.name',
            'Satellites.address',
            'Satellites.container_id',
            'Satellites.timezone',
            'Satellites.sync_method',
            'SatelliteStatus.status',
            'SatelliteStatus.last_error',
            'SatelliteStatus.last_export',
            'SatelliteStatus.last_seen',
            'SatelliteStatus.satellite_id',
            'status' => $query->newExpr('IF(SatelliteStatus.status IS NULL, 0, SatelliteStatus.status)')
        ])
            ->where($where)
            ->having($having)
            ->contain(['SatelliteStatus'])
            ->order($SatelliteFilter->getOrderForPaginator('Satellites.name', 'asc'));

        if (!empty($MY_RIGHTS)) {
            $query->andWhere([
                'Satellites.container_id IN' => $MY_RIGHTS
            ]);
        }

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    public function disableAllInstanceConfigSyncs() {
        $this->updateAll([
            'sync_instance' => 0
        ], []);
    }

    /**
     * @param array $MY_RIGHTS
     * @return \Cake\ORM\Query
     */
    public function getSatellitesForExportIndex($MY_RIGHTS = []) {
        $query = $this->find();
        $query->contain(['SatelliteStatus']);
        if (!empty($MY_RIGHTS)) {
            $query->where([
                'Satellites.container_id IN' => $MY_RIGHTS
            ]);
        }
        $query->order([
            'Satellites.name' => 'ASC'
        ])->all();

        return $query;
    }

    /**
     * @param array $containerIds
     * @param string $type
     * @return array
     */
    public function getSatellitesByContainerIds($containerIds = [], $type = 'all') {
        if (!is_array($containerIds)) {
            $containerIds = [$containerIds];
        }

        //Lookup for the tenant container of $container_id
        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        $tenantContainerIds = [];

        foreach ($containerIds as $containerId) {
            if ($containerId != ROOT_CONTAINER) {

                // Get container id of the tenant container
                // $container_id is may be a location, devicegroup or whatever, so we need to container id of the tenant container to load service templates
                $path = $ContainersTable->getPathByIdAndCacheResult($containerId, 'SatellitesByContainerId');

                // Tenant satellites are available for all users of a tenant (oITC V2 legacy)
                if (isset($path[1])) {
                    $tenantContainerIds[] = $path[1]['id'];
                }
            } else {
                $tenantContainerIds[] = ROOT_CONTAINER;
            }
        }
        $tenantContainerIds = array_unique($tenantContainerIds);
        $containerIds = array_unique(array_merge($tenantContainerIds, $containerIds));


        if (empty($containerIds)) {
            return [];
        }

        $where = [
            'Satellites.container_id IN' => $containerIds,
        ];


        $query = $this->find()
            ->contain(['Containers'])
            ->where($where)
            ->disableHydration()
            ->all();

        $records = $query->toArray();
        if (empty($records) || is_null($records)) {
            return [];
        }

        if ($type === 'all') {
            return $records;
        }

        $list = [];
        foreach ($records as $record) {
            $list[$record['id']] = $record['name'];
        }
        return $list;
    }

    /**
     * @param int $containerId
     * @param string $type
     * @param array $MY_RIGHTS
     * @param array $where
     * @return array
     */
    public function getSatellitesByContainerIdExact($containerId, $type = 'all', $index = 'id', $MY_RIGHTS = [], $where = []) {
        $_where = [
            'Satellites.container_id' => $containerId
        ];

        $where = Hash::merge($_where, $where);

        $query = $this->find();
        $query->select([
            'Satellites.' . $index,
            'Satellites.name'
        ]);
        $query->where($where);

        if (!empty($MY_RIGHTS)) {
            $query->andWhere([
                'Satellites.container_id IN' => $MY_RIGHTS
            ]);
        }

        $query->disableHydration();

        $result = $query->toArray();
        if (empty($result)) {
            return [];
        }

        if ($type === 'all') {
            return $result;
        }

        $list = [];
        foreach ($result as $row) {
            $list[$row[$index]] = $row['name'];
        }

        return $list;
    }

    /**
     * @return array
     */
    public function getSatellitesForCheckmkNagiosExportCommand() {
        $query = $this->find()
            ->where([
                'Satellites.sync_instance' => 1,
            ])->all();

        if (empty($query)) {
            return [];
        }
        return $query->toArray();
    }

    /**
     * @param string $method
     * @return array
     */
    public function getSatellitesBySyncMethod($method = 'ssh') {
        $query = $this->find()
            ->where([
                'Satellites.sync_method' => $method,
            ])->all();

        if (empty($query)) {
            return [];
        }
        return $query->toArray();
    }

    /**
     * @param int $satelliteId
     * @return array
     */
    public function getSatelliteByIdForAgent(int $satelliteId) {
        $query = $this->find()
            ->select([
                'id',
                'name',
                'address'
            ])
            ->where([
                'id' => $satelliteId
            ])
            ->disableHydration()
            ->firstOrFail();

        return $query;
    }

    /**
     * @param SystemdowntimesConditions $SystemdowntimesConditions
     * @param PaginateOMat|null $PaginateOMat
     * @return array
     */
    public function getRecurringSatelliteDowntimes(SystemdowntimesConditions $SystemdowntimesConditions, $PaginateOMat = null) {
        $MY_RIGHTS = $SystemdowntimesConditions->getContainerIds();

        $query = $this->find();
        $query->select([
            'Systemdowntimes.id',
            'Systemdowntimes.objecttype_id',
            'Systemdowntimes.object_id',
            'Systemdowntimes.downtimetype_id',
            'Systemdowntimes.weekdays',
            'Systemdowntimes.day_of_month',
            'Systemdowntimes.from_time',
            'Systemdowntimes.to_time',
            'Systemdowntimes.duration',
            'Systemdowntimes.comment',
            'Systemdowntimes.author',
            'Satellites.name',
            'Satellites.container_id'
        ])
            ->join([
                [
                    'table'      => 'systemdowntimes',
                    'alias'      => 'Systemdowntimes',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Systemdowntimes.object_id = Satellites.id'
                    ]
                ],
                [
                    'table'      => 'containers',
                    'alias'      => 'Containers',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Containers.id = Satellites.container_id'
                    ]
                ]
            ])
            ->andWhere([
                'Systemdowntimes.objecttype_id' => OBJECT_SATELLITE
            ]);

        if (!empty($MY_RIGHTS)) {
            $query->andWhere([
                'Containers.id IN' => $MY_RIGHTS
            ]);
        }

        if ($SystemdowntimesConditions->hasConditions()) {
            $query->andWhere($SystemdowntimesConditions->getConditions());
        }

        $query->group([
            'Systemdowntimes.id'
        ])
            ->order(
                array_merge(
                    $SystemdowntimesConditions->getOrder(),
                    ['Systemdowntimes.id' => 'asc']
                )
            )->disableHydration();
        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }
}
